// MiniDump01.cpp : Capture a hang dump.
//

#include "stdafx.h"
#include <windows.h>
#include <dbghelp.h>

int HangDump(HANDLE hProcess, DWORD dwProcessId, HANDLE hFile, MINIDUMP_TYPE miniDumpType);

int _tmain(int argc, TCHAR* argv[])
{
	int nResult = -1;
	HANDLE hProcess = INVALID_HANDLE_VALUE;
	DWORD dwProcessId = 0;
	HANDLE hFile = INVALID_HANDLE_VALUE;
	MINIDUMP_TYPE miniDumpType;

	// DbgHelp v5.2
	miniDumpType = (MINIDUMP_TYPE) (MiniDumpNormal | MiniDumpWithProcessThreadData | MiniDumpWithDataSegs | MiniDumpWithHandleData);
	// DbgHelp v6.3 - Passing unsupported flags to a lower version of DbgHelp does not cause any issues
	miniDumpType = (MINIDUMP_TYPE) (miniDumpType | MiniDumpWithFullMemoryInfo | MiniDumpWithThreadInfo);

	if ((argc == 2) && (_stscanf_s(argv[1], _T("%ld"), &dwProcessId) == 1))
	{
		// Generate the filename (ISO format)
		SYSTEMTIME systemTime;
		GetLocalTime(&systemTime);
		TCHAR szFilename[64];
		_stprintf_s(szFilename, 64, _T("c:\\dumps\\minidump_%04d-%02d-%02d_%02d-%02d-%02d-%03d.dmp"),
					systemTime.wYear, systemTime.wMonth, systemTime.wDay, 
					systemTime.wHour, systemTime.wMinute, systemTime.wSecond, systemTime.wMilliseconds);

		// Create the folder and file
		CreateDirectory(_T("c:\\dumps"), NULL);
		if ((hFile = CreateFile(szFilename, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL)) == INVALID_HANDLE_VALUE)
		{
			_tprintf(_T("Unable to open '%s' for write (Error: %08x)\n"), szFilename, GetLastError());
			nResult = 2;
		}
		// Open the process
		else if ((hProcess = OpenProcess(PROCESS_ALL_ACCESS, FALSE, dwProcessId)) == NULL)
		{
			_tprintf(_T("Unable to open process %ld (Error: %08x)\n"), dwProcessId, GetLastError());
			nResult = 3;
		}
		// Take a hang dump
		else
		{
			nResult = HangDump(hProcess, dwProcessId, hFile, miniDumpType);
		}
		if (hFile) CloseHandle(hFile);
		if (hProcess) CloseHandle(hProcess);
		if (nResult == 0)
		{
			_tprintf(_T("Dump Created - '%s'\n"), szFilename);
		}
		else
		{
			DeleteFile(szFilename);
		}
	}
	else
	{
		_tprintf(_T("Usage: %s <pid>\n"), argv[0]);
		nResult = 1;
	}

	return 0;
}

int HangDump(HANDLE hProcess, DWORD dwProcessId, HANDLE hFile, MINIDUMP_TYPE miniDumpType)
{
	if (!MiniDumpWriteDump(hProcess, dwProcessId, hFile, miniDumpType, NULL, NULL, NULL))
	{
		_tprintf(_T("Failed to create hang dump (Error: %08x)\n"), GetLastError());
		return 11;
	}
	return 0;
}
