// MiniDump02.cpp : Captures a hang or crash dump at runtime.
//

#include "stdafx.h"
#include <windows.h>
#include <dbghelp.h>

int WriteDump(HANDLE hProcess, DWORD dwProcessId, HANDLE hFile, MINIDUMP_TYPE miniDumpType, PMINIDUMP_EXCEPTION_INFORMATION pExceptionParam);
int DebugEventDump(HANDLE hProcess, DWORD dwProcessId, HANDLE hFile, MINIDUMP_TYPE miniDumpType);

int _tmain(int argc, TCHAR* argv[])
{
	int nResult = -1;
	HANDLE hProcess = INVALID_HANDLE_VALUE;
	DWORD dwProcessId = 0;
	ULONG64 ulJitInfoAddr = NULL;
	HANDLE hFile = INVALID_HANDLE_VALUE;
	MINIDUMP_TYPE miniDumpType;

	// DbgHelp v5.2
	miniDumpType = (MINIDUMP_TYPE) (MiniDumpNormal | MiniDumpWithProcessThreadData | MiniDumpWithDataSegs | MiniDumpWithHandleData);
	// DbgHelp v6.3 - Passing unsupported flags to a lower version of DbgHelp does not cause any issues
	miniDumpType = (MINIDUMP_TYPE) (miniDumpType | MiniDumpWithFullMemoryInfo | MiniDumpWithThreadInfo);

	if (((argc == 2) || (argc == 3)) && (_stscanf_s(argv[1], _T("%ld"), &dwProcessId) == 1))
	{
		// Generate the filename (ISO format)
		SYSTEMTIME systemTime;
		GetLocalTime(&systemTime);
		TCHAR szFilename[64];
		_stprintf_s(szFilename, 64, _T("c:\\dumps\\minidump_%04d-%02d-%02d_%02d-%02d-%02d-%03d.dmp"),
					systemTime.wYear, systemTime.wMonth, systemTime.wDay, 
					systemTime.wHour, systemTime.wMinute, systemTime.wSecond, systemTime.wMilliseconds);

		// Create the folder and file
		CreateDirectory(_T("c:\\dumps"), NULL);
		if ((hFile = CreateFile(szFilename, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL)) == INVALID_HANDLE_VALUE)
		{
			_tprintf(_T("Unable to open '%s' for write (Error: %08x)\n"), szFilename, GetLastError());
			nResult = 2;
		}
		// Open the process
		else if ((hProcess = OpenProcess(PROCESS_ALL_ACCESS, FALSE, dwProcessId)) == NULL)
		{
			_tprintf(_T("Unable to open process %ld (Error: %08x)\n"), dwProcessId, GetLastError());
			nResult = 3;
		}
		// Take a hang dump
		else if (argc == 2)
		{
			nResult = WriteDump(hProcess, dwProcessId, hFile, miniDumpType, NULL);
		}
		// Wait for a exception and take a crash dump
		else if ((argc == 3) && (_tcscmp(argv[2], _T("-e")) == 0))
		{
			nResult = DebugEventDump(hProcess, dwProcessId, hFile, miniDumpType);

			// Terminate the process
			TerminateProcess(hProcess, -1);
		}
		if (hFile) CloseHandle(hFile);
		if (hProcess) CloseHandle(hProcess);
		if (nResult == 0)
		{
			_tprintf(_T("Dump Created - '%s'\n"), szFilename);
		}
		else
		{
			DeleteFile(szFilename);
		}
	}
	else
	{
		_tprintf(_T("Usage: %s <pid>         : Hang Dump\n"), argv[0]);
		_tprintf(_T("Usage: %s <pid> -e      : Crash Dump\n"), argv[0]);
		nResult = 1;
	}

	return 0;
}

int WriteDump(HANDLE hProcess, DWORD dwProcessId, HANDLE hFile, MINIDUMP_TYPE miniDumpType, PMINIDUMP_EXCEPTION_INFORMATION pExceptionParam)
{
	if (!MiniDumpWriteDump(hProcess, dwProcessId, hFile, miniDumpType, pExceptionParam, NULL, NULL))
	{
		_tprintf(_T("Failed to create hang dump (Error: %08x)\n"), GetLastError());
		return 11;
	}
	return 0;
}

int DebugEventDump(HANDLE hProcess, DWORD dwProcessId, HANDLE hFile, MINIDUMP_TYPE miniDumpType)
{
	if (!DebugActiveProcess(dwProcessId))
		return 21;

	int n = 0;
	int nResult = -1;
	while (nResult == -1)
	{
		DEBUG_EVENT	debugEvent;
		if (WaitForDebugEvent(&debugEvent, 1000))
		{
			switch (debugEvent.dwDebugEventCode)
			{
			case EXCEPTION_DEBUG_EVENT:
				{	// Exception
					if (debugEvent.u.Exception.ExceptionRecord.ExceptionCode == EXCEPTION_BREAKPOINT)
					{	// Breakpoint; we handle it to make execution continue
						ContinueDebugEvent(debugEvent.dwProcessId, debugEvent.dwThreadId, DBG_CONTINUE);
					}
					else if (debugEvent.u.Exception.dwFirstChance != 0)
					{	// First chance; we don't handle it so the exception handlers get a go
						ContinueDebugEvent(debugEvent.dwProcessId, debugEvent.dwThreadId, DBG_EXCEPTION_NOT_HANDLED);
					}
					else
					{	// Second chance; use the ThreadId in the event to determine which thread to open to retrieve the context
						nResult = 22;
						CONTEXT context = {0};
						context.ContextFlags = CONTEXT_ALL;
						HANDLE hThread = OpenThread(THREAD_GET_CONTEXT, FALSE, debugEvent.dwThreadId);
						if ((hThread != INVALID_HANDLE_VALUE) && GetThreadContext(hThread, &context))
						{
							EXCEPTION_POINTERS exceptionPointers = {0};
							exceptionPointers.ContextRecord = &context;
							exceptionPointers.ExceptionRecord = &(debugEvent.u.Exception.ExceptionRecord);

							MINIDUMP_EXCEPTION_INFORMATION	exceptionInfo = {0};
							exceptionInfo.ThreadId = debugEvent.dwThreadId;
							exceptionInfo.ExceptionPointers = &exceptionPointers;
							exceptionInfo.ClientPointers = FALSE; // ExceptionPointers is in our address space

							nResult = WriteDump(hProcess, dwProcessId, hFile, miniDumpType, &exceptionInfo);
						}
						if (hThread) CloseHandle(hThread);
					}
				}
				break;
			case EXIT_PROCESS_DEBUG_EVENT:
				{	// Process ended before we got an exception; we handle this event here and then drop out of the loop
					ContinueDebugEvent(debugEvent.dwProcessId, debugEvent.dwThreadId, DBG_CONTINUE);
					nResult = 24;
				}
				break;
			case CREATE_PROCESS_DEBUG_EVENT:
				{	// We need to do some house-cleaning for this event...
					if (debugEvent.u.CreateProcessInfo.hFile)
						CloseHandle(debugEvent.u.CreateProcessInfo.hFile);

					ContinueDebugEvent(debugEvent.dwProcessId, debugEvent.dwThreadId, DBG_CONTINUE);
				}
				break;
			case LOAD_DLL_DEBUG_EVENT:
				{	// We need to do some house-cleaning for this event...
					if (debugEvent.u.LoadDll.hFile)
						CloseHandle(debugEvent.u.LoadDll.hFile);

					ContinueDebugEvent(debugEvent.dwProcessId, debugEvent.dwThreadId, DBG_CONTINUE);
				}
				break;
			default:
				{	// All other events are handled here
					ContinueDebugEvent(debugEvent.dwProcessId, debugEvent.dwThreadId, DBG_CONTINUE);
				}
				break;
			}
		}
		else
		{
			_tprintf(_T(".%s"), (++n % 60 == 0 ? "\n" : ""));  n = n % 60;
		}
	}
	_tprintf(_T("\n"));
	DebugActiveProcessStop(dwProcessId);
	return nResult;
}
