// MiniDump05.cpp : Captures a hang or crash dump. Uses a callback to add all committed MEM_PRIVATE and R/W memory, and survives a memory read failure.
//

#include "stdafx.h"
#include <windows.h>
#include <dbghelp.h>

int WriteDump(HANDLE hProcess, DWORD dwProcessId, HANDLE hFile, MINIDUMP_TYPE miniDumpType, PMINIDUMP_EXCEPTION_INFORMATION pExceptionParam);
int DebugEventDump(HANDLE hProcess, DWORD dwProcessId, HANDLE hFile, MINIDUMP_TYPE miniDumpType);
int JitInfoDump(HANDLE hProcess, DWORD dwProcessId, HANDLE hFile, MINIDUMP_TYPE miniDumpType, ULONG64 ulJitInfoAddr);
BOOL CALLBACK MiniDumpCallbackRoutine(__in PVOID CallbackParam, __in const PMINIDUMP_CALLBACK_INPUT CallbackInput, __inout PMINIDUMP_CALLBACK_OUTPUT CallbackOutput);

struct MemoryInfoNode
{
	MemoryInfoNode* pNext;
	ULONG64 ulBase;
	ULONG ulSize;
};

int _tmain(int argc, TCHAR* argv[])
{
	int nResult = -1;
	HANDLE hProcess = INVALID_HANDLE_VALUE;
	DWORD dwProcessId = 0;
	HANDLE hEvent = INVALID_HANDLE_VALUE;
	ULONG64 ulJitInfoAddr = NULL;
	HANDLE hFile = INVALID_HANDLE_VALUE;
	MINIDUMP_TYPE miniDumpType;

	// DbgHelp v5.2
	miniDumpType = (MINIDUMP_TYPE) (MiniDumpNormal | MiniDumpWithProcessThreadData | MiniDumpWithDataSegs | MiniDumpWithHandleData);
	// DbgHelp v6.3 - Passing unsupported flags to a lower version of DbgHelp does not cause any issues
	miniDumpType = (MINIDUMP_TYPE) (miniDumpType | MiniDumpWithFullMemoryInfo | MiniDumpWithThreadInfo);

	if (((argc == 2) || (argc == 3) || (argc == 4)) && (_stscanf_s(argv[1], _T("%ld"), &dwProcessId) == 1))
	{
		// Generate the filename (ISO format)
		SYSTEMTIME systemTime;
		GetLocalTime(&systemTime);
		TCHAR szFilename[64];
		_stprintf_s(szFilename, 64, _T("c:\\dumps\\minidump_%04d-%02d-%02d_%02d-%02d-%02d-%03d.dmp"),
					systemTime.wYear, systemTime.wMonth, systemTime.wDay, 
					systemTime.wHour, systemTime.wMinute, systemTime.wSecond, systemTime.wMilliseconds);

		// Create the folder and file
		CreateDirectory(_T("c:\\dumps"), NULL);
		if ((hFile = CreateFile(szFilename, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL)) == INVALID_HANDLE_VALUE)
		{
			_tprintf(_T("Unable to open '%s' for write (Error: %08x)\n"), szFilename, GetLastError());
			nResult = 2;
		}
		// Open the process
		else if ((hProcess = OpenProcess(PROCESS_ALL_ACCESS, FALSE, dwProcessId)) == NULL)
		{
			_tprintf(_T("Unable to open process %ld (Error: %08x)\n"), dwProcessId, GetLastError());
			nResult = 3;
		}
		// Take a hang dump
		else if (argc == 2)
		{
			nResult = WriteDump(hProcess, dwProcessId, hFile, miniDumpType, NULL);
		}
		// Wait for a exception and take a crash dump
		else if ((argc == 3) && (_tcscmp(argv[2], _T("-e")) == 0))
		{
			nResult = DebugEventDump(hProcess, dwProcessId, hFile, miniDumpType);

			// Terminate the process
			TerminateProcess(hProcess, -1);
		}
		// Post Mortem (AeDebug) dump - No JIT - pre Vista
		else if ((argc == 3) && (_stscanf_s(argv[2], _T("%ld"), &hEvent) == 1))
		{
			nResult = WriteDump(hProcess, dwProcessId, hFile, miniDumpType, NULL);

			// Finish Debugging
			SetEvent(hEvent);
		}
		// Post Mortem (AeDebug) dump - JIT_DEBUG_INFO - Vista+
		else if ((argc == 4) && (_stscanf_s(argv[2], _T("%ld"), &hEvent) == 1) && (_stscanf_s(argv[3], _T("%p"), &ulJitInfoAddr) == 1))
		{
			nResult = JitInfoDump(hProcess, dwProcessId, hFile, miniDumpType, ulJitInfoAddr);

			// Finish Debugging
			SetEvent(hEvent);
		}
		if (hFile) CloseHandle(hFile);
		if (hProcess) CloseHandle(hProcess);
		if (nResult == 0)
		{
			_tprintf(_T("Dump Created - '%s'\n"), szFilename);
		}
		else
		{
			DeleteFile(szFilename);
		}
	}
	else
	{
		_tprintf(_T("Usage: %s <pid>         : Hang Dump\n"), argv[0]);
		_tprintf(_T("Usage: %s <pid> -e      : Crash Dump\n"), argv[0]);
		_tprintf(_T("Usage: %s %%ld %%ld [%%p]   : AeDebug Dump\n"), argv[0]);
		nResult = 1;
	}

	return 0;
}

int WriteDump(HANDLE hProcess, DWORD dwProcessId, HANDLE hFile, MINIDUMP_TYPE miniDumpType, PMINIDUMP_EXCEPTION_INFORMATION pExceptionParam)
{
	MemoryInfoNode* pRootMemoryInfoNode = NULL;

	MINIDUMP_CALLBACK_INFORMATION callbackInfo;
	callbackInfo.CallbackParam = &pRootMemoryInfoNode;
	callbackInfo.CallbackRoutine = MiniDumpCallbackRoutine;

	if (!MiniDumpWriteDump(hProcess, dwProcessId, hFile, miniDumpType, pExceptionParam, NULL, &callbackInfo))
	{
		_tprintf(_T("Failed to create hang dump (Error: %08x)\n"), GetLastError());
		while (pRootMemoryInfoNode)
		{	// If there was an error, we'll need to cleanup here
			MemoryInfoNode* pNode = pRootMemoryInfoNode;
			pRootMemoryInfoNode = pNode->pNext;
			delete pNode;
		}
		return 11;
	}
	return 0;
}

int DebugEventDump(HANDLE hProcess, DWORD dwProcessId, HANDLE hFile, MINIDUMP_TYPE miniDumpType)
{
	if (!DebugActiveProcess(dwProcessId))
		return 21;

	int n = 0;
	int nResult = -1;
	while (nResult == -1)
	{
		DEBUG_EVENT	debugEvent;
		if (WaitForDebugEvent(&debugEvent, 1000))
		{
			switch (debugEvent.dwDebugEventCode)
			{
			case EXCEPTION_DEBUG_EVENT:
				{	// Exception
					if (debugEvent.u.Exception.ExceptionRecord.ExceptionCode == EXCEPTION_BREAKPOINT)
					{	// Breakpoint; we handle it to make execution continue
						ContinueDebugEvent(debugEvent.dwProcessId, debugEvent.dwThreadId, DBG_CONTINUE);
					}
					else if (debugEvent.u.Exception.dwFirstChance != 0)
					{	// First chance; we don't handle it so the exception handlers get a go
						ContinueDebugEvent(debugEvent.dwProcessId, debugEvent.dwThreadId, DBG_EXCEPTION_NOT_HANDLED);
					}
					else
					{	// Second chance; use the ThreadId in the event to determine which thread to open to retrieve the context
						nResult = 22;
						CONTEXT context = {0};
						context.ContextFlags = CONTEXT_ALL;
						HANDLE hThread = OpenThread(THREAD_GET_CONTEXT, FALSE, debugEvent.dwThreadId);
						if ((hThread != INVALID_HANDLE_VALUE) && GetThreadContext(hThread, &context))
						{
							EXCEPTION_POINTERS exceptionPointers = {0};
							exceptionPointers.ContextRecord = &context;
							exceptionPointers.ExceptionRecord = &(debugEvent.u.Exception.ExceptionRecord);

							MINIDUMP_EXCEPTION_INFORMATION	exceptionInfo = {0};
							exceptionInfo.ThreadId = debugEvent.dwThreadId;
							exceptionInfo.ExceptionPointers = &exceptionPointers;
							exceptionInfo.ClientPointers = FALSE; // ExceptionPointers is in our address space

							nResult = WriteDump(hProcess, dwProcessId, hFile, miniDumpType, &exceptionInfo);
						}
						if (hThread) CloseHandle(hThread);
					}
				}
				break;
			case EXIT_PROCESS_DEBUG_EVENT:
				{	// Process ended before we got an exception; we handle this event here and then drop out of the loop
					ContinueDebugEvent(debugEvent.dwProcessId, debugEvent.dwThreadId, DBG_CONTINUE);
					nResult = 24;
				}
				break;
			case CREATE_PROCESS_DEBUG_EVENT:
				{	// We need to do some house-cleaning for this event...
					if (debugEvent.u.CreateProcessInfo.hFile)
						CloseHandle(debugEvent.u.CreateProcessInfo.hFile);

					ContinueDebugEvent(debugEvent.dwProcessId, debugEvent.dwThreadId, DBG_CONTINUE);
				}
				break;
			case LOAD_DLL_DEBUG_EVENT:
				{	// We need to do some house-cleaning for this event...
					if (debugEvent.u.LoadDll.hFile)
						CloseHandle(debugEvent.u.LoadDll.hFile);

					ContinueDebugEvent(debugEvent.dwProcessId, debugEvent.dwThreadId, DBG_CONTINUE);
				}
				break;
			default:
				{	// All other events are handled here
					ContinueDebugEvent(debugEvent.dwProcessId, debugEvent.dwThreadId, DBG_CONTINUE);
				}
				break;
			}
		}
		else
		{
			_tprintf(_T(".%s"), (++n % 60 == 0 ? "\n" : ""));  n = n % 60;
		}
	}
	_tprintf(_T("\n"));
	DebugActiveProcessStop(dwProcessId);
	return nResult;
}

int JitInfoDump(HANDLE hProcess, DWORD dwProcessId, HANDLE hFile, MINIDUMP_TYPE miniDumpType, ULONG64 ulJitInfoAddr)
{
	JIT_DEBUG_INFO jitDebugInfo;
	MINIDUMP_EXCEPTION_INFORMATION exceptionInfo = {0};
	EXCEPTION_POINTERS exceptionPointers = {0};
	EXCEPTION_RECORD exceptionRecord = {0};
	CONTEXT contextRecord = {0};
	SIZE_T numberOfBytesRead;

	// As we are the same bitness as the target - the structure sizes will match 
    if (ReadProcessMemory(hProcess, (void*)ulJitInfoAddr, &jitDebugInfo, sizeof(jitDebugInfo), &numberOfBytesRead) && 
        (numberOfBytesRead == sizeof(jitDebugInfo)))
	{
		if (ReadProcessMemory(hProcess, (void*)jitDebugInfo.lpExceptionRecord, &exceptionRecord, sizeof(exceptionRecord), &numberOfBytesRead) &&
			(numberOfBytesRead == sizeof(exceptionRecord)))
		{
			if (ReadProcessMemory(hProcess, (void*)jitDebugInfo.lpContextRecord, &contextRecord, sizeof(contextRecord), &numberOfBytesRead) &&
				(numberOfBytesRead == sizeof(contextRecord)))
			{
				exceptionPointers.ExceptionRecord = &exceptionRecord;
				exceptionPointers.ContextRecord = &contextRecord;
				exceptionInfo.ThreadId = jitDebugInfo.dwThreadID;
				exceptionInfo.ExceptionPointers = &exceptionPointers;
				exceptionInfo.ClientPointers = FALSE; // ExceptionPointers is in our address space

				return WriteDump(hProcess, dwProcessId, hFile, miniDumpType, &exceptionInfo);
			}
		}
	}

	return WriteDump(hProcess, dwProcessId, hFile, miniDumpType, NULL);
}

BOOL CALLBACK MiniDumpCallbackRoutine(
  __in     PVOID CallbackParam,
  __in     const PMINIDUMP_CALLBACK_INPUT CallbackInput,
  __inout  PMINIDUMP_CALLBACK_OUTPUT CallbackOutput
)
{	// Callback supported in Windows 2003 SP1 unless indicated
	// Switch statement is in call order
	switch(CallbackInput->CallbackType)
	{
	case IoStartCallback:  //  DbgHelp.dll v6.5; Available in Vista/Win2008
		break;
	case SecondaryFlagsCallback:  //  DbgHelp.dll v6.6; Available in Vista/Win2008
		break;
	case CancelCallback:
		break;
	case IncludeThreadCallback:
		break;
	case IncludeModuleCallback:
		break;
	case ModuleCallback:
		break;
	case ThreadCallback:
		{	// Collect all of committed MEM_PRIVATE and R/W memory
			MemoryInfoNode** ppRoot = (MemoryInfoNode**)CallbackParam;
			if (ppRoot && !*ppRoot)	// Only do this once
			{
				MEMORY_BASIC_INFORMATION mbi;
				ULONG64 ulAddress = 0;
				SIZE_T dwSize = VirtualQueryEx(CallbackInput->ProcessHandle, (void*)ulAddress, &mbi, sizeof(MEMORY_BASIC_INFORMATION));
				while (dwSize == sizeof(MEMORY_BASIC_INFORMATION))
				{
					if ((mbi.State == MEM_COMMIT) &&
						((mbi.Type == MEM_PRIVATE) || (mbi.Protect == PAGE_READWRITE) || (mbi.Protect == PAGE_EXECUTE_READWRITE)))
					{
						MemoryInfoNode* pNode = new MemoryInfoNode;
						pNode->ulBase = (ULONG64)mbi.BaseAddress;
						pNode->ulSize = (ULONG)mbi.RegionSize;
						pNode->pNext = *ppRoot;
						*ppRoot = pNode;
					}
					// Loop
					ulAddress = (ULONG64)mbi.BaseAddress + mbi.RegionSize;
					dwSize = VirtualQueryEx(CallbackInput->ProcessHandle, (void*)ulAddress, &mbi, sizeof(MEMORY_BASIC_INFORMATION));
				}
			}
		}
		break;
	case ThreadExCallback:
		break;
	case MemoryCallback:
		{	// Remove the root node and return its members in the callback
			MemoryInfoNode** ppRoot = (MemoryInfoNode**)CallbackParam;
			if (ppRoot && *ppRoot)
			{
				MemoryInfoNode* pNode = *ppRoot;
				*ppRoot = pNode->pNext;
				CallbackOutput->MemoryBase = pNode->ulBase;
				CallbackOutput->MemorySize = pNode->ulSize;
				delete pNode;
			}
		}
		break;
	case RemoveMemoryCallback:
		break;
	case WriteKernelMinidumpCallback:
		break;
	case KernelMinidumpStatusCallback:
		break;
	case IncludeVmRegionCallback:  //  DbgHelp.dll v6.5; Available in Vista/Win2008; Requires MiniDumpWithFullMemory
		break;
	case IoWriteAllCallback:  //  DbgHelp.dll v6.5; Available in Vista/Win2008
		break;
	case IoFinishCallback:  //  DbgHelp.dll v6.5; Available in Vista/Win2008
		break;
	case ReadMemoryFailureCallback:  // DbgHelp.dll v6.5; Available in Vista/Win2008
		{	//  There has been a failure to read memory. Set Status to S_OK to ignore it.
			CallbackOutput->Status = S_OK;
		}
		break;
	}
	return TRUE;
}
