// MiniDump06.cpp : Export a callback to add all committed MEM_PRIVATE and R/W memory, and survives a memory read failure.
//

#include "stdafx.h"
#include <windows.h>
#include <dbghelp.h>

struct MemoryInfoNode
{
	MemoryInfoNode* pNext;
	ULONG64 ulBase;
	ULONG ulSize;
};

MemoryInfoNode* g_pRootMemoryInfoNode = NULL;

BOOL CALLBACK MiniDumpCallbackRoutine(
  __in     PVOID CallbackParam,
  __in     const PMINIDUMP_CALLBACK_INPUT CallbackInput,
  __inout  PMINIDUMP_CALLBACK_OUTPUT CallbackOutput
)
{	// Callback supported in Windows 2003 SP1 unless indicated
	// Switch statement is in call order
	switch(CallbackInput->CallbackType)
	{
	case IoStartCallback:  //  DbgHelp.dll v6.5; Available in Vista/Win2008
		break;
	case SecondaryFlagsCallback:  //  DbgHelp.dll v6.6; Available in Vista/Win2008
		break;
	case CancelCallback:
		break;
	case IncludeThreadCallback:
		{
			while (g_pRootMemoryInfoNode)
			{	// Unexpected cleanup required
				MemoryInfoNode* pNode = g_pRootMemoryInfoNode;
				g_pRootMemoryInfoNode = pNode->pNext;
				delete pNode;
			}
		}
		break;
	case IncludeModuleCallback:
		break;
	case ModuleCallback:
		break;
	case ThreadCallback:
		{	// Collect all of committed MEM_PRIVATE and R/W memory
			if (!g_pRootMemoryInfoNode)	// Only do this once
			{
				MEMORY_BASIC_INFORMATION mbi;
				ULONG64 ulAddress = 0;
				SIZE_T dwSize = VirtualQueryEx(CallbackInput->ProcessHandle, (void*)ulAddress, &mbi, sizeof(MEMORY_BASIC_INFORMATION));
				while (dwSize == sizeof(MEMORY_BASIC_INFORMATION))
				{
					if ((mbi.State == MEM_COMMIT) &&
						((mbi.Type == MEM_PRIVATE) || (mbi.Protect == PAGE_READWRITE) || (mbi.Protect == PAGE_EXECUTE_READWRITE)))
					{
						MemoryInfoNode* pNode = new MemoryInfoNode;
						pNode->ulBase = (ULONG64)mbi.BaseAddress;
						pNode->ulSize = (ULONG)mbi.RegionSize;
						pNode->pNext = g_pRootMemoryInfoNode;
						g_pRootMemoryInfoNode = pNode;
					}
					// Loop
					ulAddress = (ULONG64)mbi.BaseAddress + mbi.RegionSize;
					dwSize = VirtualQueryEx(CallbackInput->ProcessHandle, (void*)ulAddress, &mbi, sizeof(MEMORY_BASIC_INFORMATION));
				}
			}
		}
		break;
	case ThreadExCallback:
		break;
	case MemoryCallback:
		{	// Remove the root node and return its members in the callback
			if (g_pRootMemoryInfoNode)
			{
				MemoryInfoNode* pNode = g_pRootMemoryInfoNode;
				g_pRootMemoryInfoNode = pNode->pNext;
				CallbackOutput->MemoryBase = pNode->ulBase;
				CallbackOutput->MemorySize = pNode->ulSize;
				delete pNode;
			}
		}
		break;
	case RemoveMemoryCallback:
		break;
	case WriteKernelMinidumpCallback:
		break;
	case KernelMinidumpStatusCallback:
		break;
	case IncludeVmRegionCallback:  //  DbgHelp.dll v6.5; Available in Vista/Win2008; Requires MiniDumpWithFullMemory
		break;
	case IoWriteAllCallback:  //  DbgHelp.dll v6.5; Available in Vista/Win2008
		break;
	case IoFinishCallback:  //  DbgHelp.dll v6.5; Available in Vista/Win2008
		break;
	case ReadMemoryFailureCallback:  // DbgHelp.dll v6.5; Available in Vista/Win2008
		{	//  There has been a failure to read memory. Set Status to S_OK to ignore it.
			CallbackOutput->Status = S_OK;
		}
		break;
	}
	return TRUE;
}
